{*******************************************************}
{                                                       }
{       ReportBuilder 11 Wrapper for RichView Package   }
{       PDF renderer.                                   }
{                                                       }
{       ReportBuilder                                   }
{       Copyright (c) Digital Metaphors Corporation     }
{                                                       }
{       RichView                                        }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

{
 Include this unit in your project to provide a high quality
 PDF export of ppRichView (otherwise they will be exported as
 a bitmap image)
}

{ REPORTBUILDER14 must be defined for ReportBuilder v14 or newer. }
{$DEFINE REPORTBUILDER14}
{ REPORTBUILDER15 must be defined for ReportBuilder v15 or newer.
  For older versions, remove this define (from the line below and from ppRichView.pas }
{$DEFINE REPORTBUILDER15}
unit ppPDFRendererRV;

interface

uses
  Classes,
  ppDevice,
  RVClasses,
  ppPDFRenderer, ppPDFRendererGeneric,
  ppRichView;

type
  TppPDFRendererRichView = class(TppPDFRendererGeneric)
  private
    FRVCount: Integer;
  public
    class function DrawCommandClass: TppDrawCommandClass; override;
    procedure RenderToPDF; override;
  end;

implementation

uses
  Graphics,
  SysUtils,
  ppDrwCmd,
  {$IFDEF REPORTBUILDER15}
  ppFilDev,
  {$ENDIF}
  ppPDFRendererImage,
  ppPDFRendererManager,
  ppPDFUtils;

{ TppPDFRendererRichView }

{------------------------------------------------------------------------------}
class function TppPDFRendererRichView.DrawCommandClass: TppDrawCommandClass;
begin
  Result := TppDrawRichView;
end;

{------------------------------------------------------------------------------}
procedure TppPDFRendererRichView.RenderToPDF;
{$IFNDEF REPORTBUILDER15}
var
  lMetaFile:  TMetaFile;
  lDrawImage: TppDrawImage;
  lRenderer:  TppPDFRendererImage;
begin
  if not(DrawCommand is TppDrawRichView) then
    exit;

  lDrawImage := TppDrawImage.Create(nil);
  lRenderer := TppPDFRendererImage.Create;
  lMetaFile := TppDrawRichView(DrawCommand).GetAsMetafile(False);

  // lMetaFile.SaveToFile('d:\test\1.wmf');

  lDrawImage.Picture.MetaFile := lMetaFile;
  inc(FRVCount);
  lDrawImage.ImageName := 'ppRichView' + IntToStr(FRVCount);
  lDrawImage.Width := DrawCommand.Width;
  lDrawImage.Height := DrawCommand.Height;
  lDrawImage.Top := DrawCommand.Top;
  lDrawImage.Left := DrawCommand.Left;
  lDrawImage.Stretch := True;

  lRenderer.DrawCommand := lDrawImage;
  lRenderer.MemoryStream := MemoryStream;
  lRenderer.PDFSettings := PDFSettings;
  lRenderer.PDFXRef := PDFXRef;
  lRenderer.PageHeight := PageHeight;
  lRenderer.Printer := Printer;
  {$IFNDEF REPORTBUILDER14}
  lRenderer.PDFCanvas.InitOrigin := False;
  {$ENDIF}
  lRenderer.RenderToPDF;
  lDrawImage.Free;
  lRenderer.Free;
  lMetaFile.Free;
end;
{$ELSE}

var
  lBitmap:    TBitmap;
  lMetaFile:  TMetaFile;
  lDrawImage: TppDrawImage;
  lRenderer:  TppPDFRendererImage;
begin
  if not(DrawCommand is TppDrawRichView) then
    exit;

  lBitmap := nil;
  lMetaFile := DrawCommand.AsMetafile;

  lDrawImage := TppDrawImage.Create(nil);
  lRenderer := TppPDFRendererImage.Create;
  PDFSettings.CanvasInitOrigin := False;

  if PDFSettings.ExportRTFAsBitmap then
    lRenderer.ForceImage := True;

  if (lMetaFile = nil) or ((lRenderer.ForceImage) and
    (PDFSettings.MetafileToBitmapQuality = 1)) then
    lBitmap := TppFileDeviceUtils.DrawCommandToBitmap(DrawCommand,
      (PDFSettings.MetafileToBitmapQuality * 100), clWhite);

  try
    if lBitmap <> nil then
      lDrawImage.Picture.Bitmap := lBitmap
    else if lMetaFile <> nil then
      lDrawImage.Picture.MetaFile := lMetaFile;

    // lMetaFile.SaveToFile('d:\test\1.wmf');

    lDrawImage.ImageName := 'ppGeneric' + IntToStr(FRVCount);
    lDrawImage.Width := DrawCommand.Width;
    lDrawImage.Height := DrawCommand.Height;
    lDrawImage.Top := DrawCommand.Top;
    lDrawImage.Left := DrawCommand.Left;
    lDrawImage.Border := DrawCommand.Border;
    lDrawImage.Stretch := True;

    {Draw image to PDF}
    if (lRenderer <> nil) then
    begin
      lRenderer.DrawCommand := lDrawImage;
      lRenderer.MemoryStream := MemoryStream;
      lRenderer.PDFSettings := PDFSettings;
      lRenderer.PDFXRef := PDFXRef;
      lRenderer.PageHeight := PageHeight;
      lRenderer.Printer := Printer;
      lRenderer.GraphicsCanvas := GraphicsCanvas;
      lRenderer.RenderToPDF;
    end;

  finally
    lDrawImage.Free;
    lRenderer.Free;

    if lBitmap <> nil then
      lBitmap.Free
    else if lMetaFile <> nil then
      lMetaFile.Free;
  end;

end;
{$ENDIF}

initialization

TppPDFRendererManager.RegisterRenderer(TppPDFRendererRichView);

finalization

TppPDFRendererManager.UnregisterRenderer(TppPDFRendererRichView);

end.
