unit ppRTFRendererRV;

interface

{$I RV_Defs.inc}

uses
  Windows, Graphics, Types, Classes,
  ppRTFRenderer, ppDevice, ppDrwCmd,
  RVFuncs, RVClasses, RVStyle, RVReport, RVTypes, RVScroll, RichView, RVItem,
  RVStr,
  ppRichView;

type

  TppRTFRendererRichView = class(TppRTFRenderer)
  private
    function GetDrawRVCommand: TppDrawRichView;
    procedure GetRTFDoc(var Res, Img: String; var MarginsEMU: TRect;
      var BackStyle: TBackgroundStyle);
  public
    class function DrawCommandClass: TppDrawCommandClass; override;
    procedure RenderToRTF; override;
  end;

  TppRVRTFTables = class(TCustomRVRTFTables)
  private
    FRenderer: TppRTFRendererRichView;
  public
    constructor Create(Renderer: TppRTFRendererRichView);
    procedure ClearColors; override;
    procedure ClearFonts; override;
    procedure AddColor(Color: TColor); override;
    function GetColorIndex(Color: TColor): Integer; override;
    function AddFontFromStyle(TextStyle: TCustomRVFontInfo): Integer; override;
    function AddFontFromStyleEx(TextStyle: TCustomRVFontInfo;
      ValidProperties: TRVFontInfoProperties; Root: Boolean): Integer; override;
    function AddFont(Font: TFont): Integer; override;
    function GetFontStyleIndex(TextStyle: TCustomRVFontInfo): Integer; override;
    function GetFontIndex(Font: TFont): Integer; override;
  end;

implementation

uses
  Forms,
  SysUtils,
  ExtCtrls,
  ppTypes,
  ppRTFRendererManager,
  ppFilDev;

class function TppRTFRendererRichView.DrawCommandClass: TppDrawCommandClass;
begin
  Result := TppDrawRichView;
end;

{------------------------------------------------------------------------------}
procedure TppRTFRendererRichView.GetRTFDoc(var Res, Img: String;
  var MarginsEMU: TRect; var BackStyle: TBackgroundStyle);
var
  Stream:      TRVMemoryStream;
  Helper:      TRVReportHelper;
  RVStyle:     TRVStyle;
  DrawCommand: TppDrawRichView;
  RTFTables:   TppRVRTFTables;
  StyleToFont: TRVIntegerList;
  i:           Integer;
  s:           TRVAnsiString;
  SavingData:  TRVRTFSavingData;
begin
  Img := '';
  DrawCommand := GetDrawRVCommand;
  DrawCommand.ValidateCanvasRef;
  Stream := TRVMemoryStream.Create;
  try
    Helper := TRVReportHelperRef.Create(nil);
    RVStyle := TRVStyle.Create(nil);
    try
      RVStyle.TextStyles.PixelsPerInch := DrawCommand.InternalPixelsPerInch;
      TReportRichViewRef(Helper.RichView).FInternalPixelsPerInch :=
        DrawCommand.InternalPixelsPerInch;
      if DrawCommand.InternalPixelsPerInch > 0 then
        DrawCommand.CanvasRef.Font.PixelsPerInch :=
          DrawCommand.InternalPixelsPerInch;
      RVSetHelperDefaults(Helper);
      Helper.RichView.Style := RVStyle;
      {$IFNDEF RVDONOTUSESTYLETEMPLATES}
      Helper.RichView.UseStyleTemplates := False;
      {$ENDIF}
      Stream.Position := 0;
      DrawCommand.RVFStream.Position := 0;
      Helper.RichView.LoadRVFFromStream(DrawCommand.RVFStream);
      Helper.Init(DrawCommand.CanvasRef, DrawCommand.PageWidth);
      Helper.RichView.RTFOptions := [rvrtfDuplicateUnicode, rvrtfSaveEMFAsWMF,
        rvrtfSaveBitmapDefault];

      RTFTables := TppRVRTFTables.Create(Self);
      StyleToFont := TRVIntegerList.Create;
      try
        for i := 0 to RVStyle.TextStyles.Count - 1 do
          StyleToFont.Add(RTFTables.AddFontFromStyle(RVStyle.TextStyles[i]));
        Helper.RichView.RVData.MakeRTFTables(RTFTables, nil, False);
        Helper.RichView.RVData.FillRTFTables(RTFTables, nil);
        FillChar(SavingData, sizeof(TRVRTFSavingData), 0);
        SavingData.RTFTables := RTFTables;
        SavingData.StyleToFont := StyleToFont;
        Helper.RichView.RVData.SaveRTFToStream(Stream, False, 0, clNone, nil,
          SavingData, rvrtfdtMain, False, False, nil, nil, nil, nil, nil, nil);
      finally
        RTFTables.Free;
        StyleToFont.Free;
      end;
      SetLength(s, Stream.Size);
      Stream.Position := 0;
      Stream.ReadBuffer(PRVAnsiChar(s)^, Length(s));
      Res := String(s);
      MarginsEMU.Left := RVStyle.PixelsToTwips
        (Helper.RichView.LeftMargin) * 635;
      MarginsEMU.Top := RVStyle.PixelsToTwips(Helper.RichView.TopMargin) * 635;
      MarginsEMU.Right := RVStyle.PixelsToTwips
        (Helper.RichView.RightMargin) * 635;
      MarginsEMU.Bottom := RVStyle.PixelsToTwips
        (Helper.RichView.BottomMargin) * 635;
      BackStyle := Helper.RichView.BackgroundStyle;
      if (BackStyle <> bsNoBitmap) and not Helper.RichView.BackgroundBitmap.Empty
      then
      begin
        Stream.Clear;
        RVSaveImageToRTF(Stream, RVStyle, Helper.RichView.BackgroundBitmap, 0,
          0, Helper.RichView.RTFOptions, nil);
        Stream.Position := 0;
        SetLength(s, Stream.Size);
        Stream.ReadBuffer(PRVAnsiChar(s)^, Length(s));
        Img := String(s);
      end;
    finally
      Helper.Free;
      RVStyle.Free;
    end;
  finally
    Stream.Free;
  end;
end;

{------------------------------------------------------------------------------}
function TppRTFRendererRichView.GetDrawRVCommand: TppDrawRichView;
begin
  Result := TppDrawRichView(RTFObject.DrawCommand);
end;

{------------------------------------------------------------------------------}
procedure TppRTFRendererRichView.RenderToRTF;
var
  DrawCommand:             TppDrawRichView;
  Margins:                 TRect;
  RTFString, RTFImgString: String;
  BackStyle:               TBackgroundStyle;
begin
  if not(ecRichText in RTFSettings.ExportComponents) then
    exit;
  DrawCommand := GetDrawRVCommand;
  WriteMem('{\shp{\*\shpinst\shpbxpage\shpbypage');
  WriteMem('\shpleft' + IntToStr(TppFileDeviceUtils.MicronsToTwips
    (DrawCommand.Left)));
  WriteMem('\shptop' + IntToStr(TppFileDeviceUtils.MicronsToTwips
    (DrawCommand.Top)));
  WriteMem('\shpright' + IntToStr(TppFileDeviceUtils.MicronsToTwips
    (DrawCommand.Left + DrawCommand.Width)));
  WriteMem('\shpbottom' + IntToStr(TppFileDeviceUtils.MicronsToTwips
    (DrawCommand.Top + DrawCommand.Height)));
  WriteMem('\shpwr3\shpwrk0\shpfblwtxt0'); // text wrapping options
  WriteMem('\shpfhdr0'); // shape is in the main document
  WriteMem('\shpz' + IntToStr(ID));   // z-order
  WriteMem('\shplid' + IntToStr(ID)); // unique id
  WriteMem('{\sp{\sn fBehindDocument}{\sv 1}}');
  // place the shape behind the text
  WriteMem('{\sp{\sn shapeType}{\sv 202}}'); // Shape Type - text box
  WriteMem('{\sp{\sn fLine}{\sv 0}}'); // No border
  if (DrawCommand.Color <> clNone) and not DrawCommand.Transparent then
    WriteMem('{\sp{\sn fillColor}{\sv ' + IntToStr(DrawCommand.Color) + '}}');
  // Fill color
  GetRTFDoc(RTFString, RTFImgString, Margins, BackStyle);
  WriteMem('{\sp{\sn dyTextTop}{\sv ' + IntToStr(Margins.Top) + '}}');
  // Top margin
  WriteMem('{\sp{\sn dyTextBottom}{\sv ' + IntToStr(Margins.Bottom) + '}}');
  // Bottom margin
  WriteMem('{\sp{\sn dxTextLeft}{\sv ' + IntToStr(Margins.Left) + '}}');
  // Left margin
  WriteMem('{\sp{\sn dxTextRight}{\sv ' + IntToStr(Margins.Right) + '}}');
  // Right margin
  if RTFImgString <> '' then
  begin
    if BackStyle in [bsTiledAndScrolled, bsTiled] then
      WriteMem('{\sp{\sn fillType}{\sv 2}}') // Tiled picture
    else
      WriteMem('{\sp{\sn fillType}{\sv 3}}'); // Scaled picture

    WriteMem('{\sp{\sn fFilled}{\sv 1}}{\sp{\sn fillShape}{\sv 1}}{\sp{\sn fillBlip}{\sv '
      + RTFImgString + '}}'); // Background picture
  end
  else if (DrawCommand.Color = clNone) or DrawCommand.Transparent then
    WriteMem('{\sp{\sn fFilled}{\sv 0}}'); // Transparent
  WriteMem('{\shptxt\uc1 ' + RTFString + '}');
  // Content (\uc1 is because rvrtfDuplicateUnicode in RTFOptions)
  WriteLine('}}');

end;

{================================= TppRVRTFTables =============================}
constructor TppRVRTFTables.Create(Renderer: TppRTFRendererRichView);
begin
  inherited Create;
  FRenderer := Renderer;
end;

{------------------------------------------------------------------------------}
procedure TppRVRTFTables.ClearColors;
begin
  // do nothing
end;

{------------------------------------------------------------------------------}
procedure TppRVRTFTables.ClearFonts;
begin
  // do nothing
end;

{------------------------------------------------------------------------------}
procedure TppRVRTFTables.AddColor(Color: TColor);
begin
  GetColorIndex(Color);
end;

{------------------------------------------------------------------------------}
function TppRVRTFTables.GetColorIndex(Color: TColor): Integer;
begin
  if Color = clNone then
    Result := 0
  else
    Result := FRenderer.ColorTable.Add(ColorToRGB(Color)) + 1;
end;

{------------------------------------------------------------------------------}
function TppRVRTFTables.AddFont(Font: TFont): Integer;
begin
  Result := GetFontIndex(Font);
end;

{------------------------------------------------------------------------------}
function TppRVRTFTables.AddFontFromStyle(TextStyle: TCustomRVFontInfo): Integer;
begin
  Result := GetFontStyleIndex(TextStyle);
end;

{------------------------------------------------------------------------------}
function TppRVRTFTables.AddFontFromStyleEx(TextStyle: TCustomRVFontInfo;
  ValidProperties: TRVFontInfoProperties; Root: Boolean): Integer;
begin
  raise ERichViewError.Create('This method must not be called in this context');
end;

{------------------------------------------------------------------------------}
function TppRVRTFTables.GetFontIndex(Font: TFont): Integer;
begin
  Result := FRenderer.FontTable.Add(Font);
end;

{------------------------------------------------------------------------------}
function TppRVRTFTables.GetFontStyleIndex(TextStyle: TCustomRVFontInfo)
  : Integer;
var
  Font: TFont;
begin
  Font := TFont.Create;
  try
    Font.Assign(TextStyle);
    Result := GetFontIndex(Font);
  finally
    Font.Free;
  end;
end;

initialization

TppRTFRendererManager.RegisterRenderer(TppRTFRendererRichView);

finalization

TppRTFRendererManager.UnRegisterRenderer(TppRTFRendererRichView);

end.
