{*******************************************************}
{                                                       }
{       ReportBuilder 9-16 Wrapper for RichView Package }
{       Main unit                                       }
{                                                       }
{       ReportBuilder                                   }
{       Copyright (c) Digital Metaphors Corporation     }
{                                                       }
{       TRichView                                       }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

{

 Add this unit to the 'uses' clauses of the main form of your application to
 use it.

 Main public properties:
 (read-only) Helper: TRVReportHelper
 (read-only) RichView: TRichView (shortcut to Helper.RichView).
 Use them to assign events.
 If you change a contents of RichView by code, call ppCustomRichView.Reset
 afterward.
}

{$J+}
{$B-}
{$R ppRichView}
(*------------------------------------------------------------------------------
 Conditional defines (remove dot before $DEFINE to activate):

 RV_USE_PRAGNAAN - enable support for Pragnaan's ExportDevices (www.pragnaan.com)
 RV_PRAGNAAN_PRO - enable support for Pragnaan's ExportDevices PRO
 (RV_USE_PRAGNAAN must be enabled)

 RV_USE_HTML - use RVHtmlImporter (www.trichview.com/resources/) for reading
 from database (for TppDBRichView). At first, TppDBRichView tries to read
 RVF. If failed, it reads RTF. If failed, it reads either text or HTML,
 depending on this define.

 Support for Waler's ExtraDevices (www.waler.com) does not require conditional
 defines in ppRichView.pas, but requires conditional defines in ExtraDevices
 code (uncomment {$DEFINE USERICHVIEW} in TXDEFS.pas; if there is no such
 define there, contact the ExtraDevices developer for the updated version)

 ------------------------------------------------------------------------------*)

{$I RV_Defs.inc}
{$DEFINE REPORTBUILDER10} // - ReportBuilder 10 or newer
{$DEFINE REPORTBUILDER12} // - ReportBuilder 12 or newer
{$DEFINE REPORTBUILDER15} // - ReportBuilder 15 or newer
{ REPORTBUILDER15 must be defined for ReportBuilder v15 or newer.
  For older versions, remove this define (from the line below and from ppPDFRendererRV.pas }

{.$DEFINE RV_USE_PRAGNAAN}
{.$DEFINE RV_PRAGNAAN_PRO}
{.$DEFINE RV_USE_HTML}
{.$DEFINE RV_USE_LLPDFLIB}

{$DEFINE RVALOWEMPTYDBRICHVIEW}

unit ppRichView;

interface

uses

  {$IFDEF WIN32}
  Windows,
  {$ELSE}
  WinTypes, WinProcs,
  {$ENDIF}
  SysUtils, Classes, Graphics, Controls, Forms, StdCtrls, ExtCtrls, Menus,
  ComCtrls, Dialogs,
  ppClass, ppTypes, ppUtils, ppCtrls,
  ppDB, ppDevice,
  ppPrintr, ppStrtch, ppViewr, ppDrwCmd, ppRTTI, ppComm,
  {$IFDEF REPORTBUILDER12}
  ppGraphicsDevice,
  {$ENDIF}
  {$IFDEF RV_USE_HTML}
  RVHtmlImport,
  {$ENDIF}
  {$IFDEF RV_USE_PRAGNAAN}
  PsRBExport_Main, PsRBExport_Graphic, PsRBExport_PDF, PsRBExport_RTF,
  PsRBExport_Excel,
  {$IFDEF RV_PRAGNAAN_PRO}
  PsRBExport_Utils,
  {$ENDIF}
  {$ENDIF}
  CRVData, RVClasses, RVRVData, RVFuncs, RichEdit, RichView, RVStyle, RVReport,
  PtblRV,
  RVStr, RVTypes;

const

  TPPRICHVIEWREFCANVASPPI  = 1440;
  TPPRICHVIEWUSEPRINTERRES = False;

  MINPPRICHVIEWHEIGHTMM = 0;

type

  TRVReportHelperRef = class(TRVReportHelper)
  private
    FFormattedPageCount: Integer;
  protected
    function CreateRichView: TCustomPrintableRV; override;
    {$IFNDEF RVDONOTUSESEQ}
    function IgnoreEndnotes: Boolean; override;
    {$ENDIF}
  public
    procedure Init(ACanvas: TCanvas; APageWidth: Integer);
    procedure Reset;
    procedure Clear;
    function FormatNextPage(AMaxHeight: Integer): Boolean;
  end;

  TReportRichViewRef = class(TReportRichView)
  protected
    function GetDataClass: TRichViewRVDataClass; override;
  public
    FInternalPixelsPerInch: Integer;
    constructor Create(AOwner: TComponent); override;
  end;

  { TppCustomRichViewSaveRec - page cache info }
  TppCustomRichViewSaveRec = record
    FDrawComponent: Boolean;
    FDrawShifter: Boolean;
    FKeptTogether: Boolean;
    FOverFlow: Boolean;
    FObjectIndex: Integer;
    FPartialGeneration: Boolean;

    FPageNo: Integer;
    FPageHeight, FFirstPageHeight: Integer;
  end;

  TppCustomRichView = class(TppStretchable)
  private
    FHelper:    TRVReportHelperRef;
    FCanvasRef: TCanvas;
    FRVStyle:   TRVStyle;
    FCurrentPageNo, FFirstPageHeight, FCurrentPageHeight: Integer;
    FRequiresReformatting:           Boolean;
    FInternalPixelsPerInch:          Integer;
    FUnicode, FUsePrinterResolution: Boolean;
    function GetRichView: TCustomRichView;
    function GetHelper: TRVReportHelper;
    procedure SetInternalPixelsPerInch(const Value: Integer);
    procedure SetUnicode(const Value: Boolean);
    function GetLPX: Integer;
    function GetLPY: Integer;
  protected
    procedure BoundsChange; override;
    procedure PropertiesToDrawCommand(aDrawCommand: TppDrawCommand); override;
    procedure UpdateDocument; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CalcSpaceUsed; override;
    function HasFont: Boolean; override;

    procedure RestoreFromCache(aCachePageNo: Longint); override;
    procedure SaveToCache(aCachePageNo: Longint); override;

    procedure Clear; override;
    procedure Init; override;
    procedure InitHelper;

    property Height;
    property Left;
    property Stretch;
    property Top;
    property UserName;
    property Visible;
    property Width;

    property RichView: TCustomRichView read GetRichView;
    property Helper: TRVReportHelper read GetHelper;
    property InternalPixelsPerInch: Integer read FInternalPixelsPerInch
      write SetInternalPixelsPerInch default TPPRICHVIEWREFCANVASPPI;
    property Unicode: Boolean read FUnicode write SetUnicode default False;
    property UsePrinterResolution: Boolean read FUsePrinterResolution
      write FUsePrinterResolution default TPPRICHVIEWUSEPRINTERRES;
  end;

  TppRichView = class(TppCustomRichView)
  public
    constructor Create(AOwner: TComponent); override;
  protected
    procedure DefineProperties(Filer: TFiler); override;
  published
    {$IFDEF REPORTBUILDER10}
    property Anchors;
    {$ENDIF}
    property BottomOffset;
    property Caption;
    property Color;
    property Height;
    property Left;
    property OnDrawCommandClick;
    property OnDrawCommandCreate;
    property OverFlowOffset;
    property ReprintOnOverFlow;
    property ShiftRelativeTo;
    property ShiftWithParent;
    property Stretch;
    property Transparent;
    property Top;
    property UserName;
    property Visible;
    property Width;
    property InternalPixelsPerInch;
    property Unicode;
    property KeepTogether;
  end;

  TppDBRichView = class(TppCustomRichView)
  private
    FAutoDisplay:    Boolean;
    FCurRecordNo:    Integer;
    FOnLoadDocument: TNotifyEvent;
    FOnGetRichText:  TppGetTextEvent;
    procedure SetAutoDisplay(aValue: Boolean);
    procedure Load(Stream: TStream; Unicode: Boolean);
    procedure LoadFromField(Updating: Boolean);
  protected
    procedure SetDataField(const Value: String); override;
    procedure UpdateDocument; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Notify(aCommunicator: TppCommunicator;
      aOperation: TppOperationType); override;
    function IsDataAware: Boolean; override;
    procedure CalcSpaceUsed; override;
    procedure Init; override;
  published
    {$IFDEF REPORTBUILDER10}
    property Anchors;
    {$ENDIF}
    property AutoDisplay: Boolean read FAutoDisplay write SetAutoDisplay
      default True;
    property BottomOffset;
    property Caption;
    property Color;
    property DataField;
    property DataPipeline;
    property Height;
    property Left;
    property OnDrawCommandClick;
    property OnDrawCommandCreate;
    property OverFlowOffset;
    property ParentDataPipeline;
    property ShiftRelativeTo;
    property ShiftWithParent;
    property Stretch;
    property Top;
    property Transparent;
    property UserName;
    property Visible;
    property Width;
    property InternalPixelsPerInch;
    property Unicode;
    property OnLoadDocument: TNotifyEvent read FOnLoadDocument
      write FOnLoadDocument;
    property OnGetRichText: TppGetTextEvent read FOnGetRichText
      write FOnGetRichText;
    property KeepTogether;
  end;

  TppDrawRichView = class(TppDrawCommand)
  private
    procedure StreamWriter(Stream: TStream);
    procedure StreamReader(Stream: TStream);
  protected
    FFreeCanvas:             Boolean;
    FColor:                  TColor;
    FTransparent:            Boolean;
    FPageWidth, FPageHeight: Integer;
    FCanvasRef:              TCanvas;
    FStream:                 TMemoryStream;
    FInternalPixelsPerInch, FPrinterPixelsPerInch: Integer;
    function DoDraw(FHelper: TRVReportHelper; aDevice: TppDevice): Boolean;
    procedure DefineProperties(Filer: TFiler); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    function Draw(aDevice: TppDevice): Boolean; override;
    procedure DrawTo(lCanvas: TCanvas; ADrawRect: TRect;
      MetafileCompatible, AllowMetafiles: Boolean);
    procedure SaveToStream(AStream: TStream);

    function GetAsMetafile(AllowMetafilesInside: Boolean): TMetafile;
    procedure ValidateCanvasRef;

    {$IFDEF REPORTBUILDER15}
    function AsMetaFile: TMetafile; override;
    {$ENDIF}
    property RVFStream: TMemoryStream read FStream;
    property CanvasRef: TCanvas read FCanvasRef;
  published
    property DataType default dtBlob;
    property Color:       TColor read FColor write FColor;
    property Transparent: Boolean read FTransparent write FTransparent
      default False;
    property PageHeight:            Integer read FPageHeight write FPageHeight;
    property PageWidth:             Integer read FPageWidth write FPageWidth;
    property InternalPixelsPerInch: Integer read FInternalPixelsPerInch
      write FInternalPixelsPerInch default 0;
    property PrinterPixelsPerInch: Integer read FPrinterPixelsPerInch
      write FPrinterPixelsPerInch default 0;
  end;

  TraTppCustomRichViewRTTI = class(TraTppStretchableRTTI)
  public
    class procedure GetPropList(aClass: TClass;
      aPropList: TraPropList); override;
    class function GetPropRec(aClass: TClass; const aPropName: String;
      var aPropRec: TraPropRec): Boolean; override;
    class function GetPropValue(aObject: TObject; const aPropName: String;
      var aValue): Boolean; override;
    class function RefClass: TClass; override;
    class function SetPropValue(aObject: TObject; const aPropName: String;
      var aValue): Boolean; override;
    class function CallMethod(aObject: TObject; const aMethodName: String;
      aParams: TraParamList; aGet: Boolean): Boolean; override;
    class function GetParams(const aMethodName: String): TraParamList; override;
  end;

  {$IFDEF RV_USE_PRAGNAAN}

var
  RVppsInRTFAsImages: Boolean = False;
  {$ENDIF}
  {$IFDEF RV_USE_HTML}

var
  ppExternalHtmlImporter: TRVHTMLImporter = nil;
  {$ENDIF}
procedure RVSetHelperDefaults(helper: TRVReportHelper);

implementation

uses
  ppBands, ppRegion, ppPrnDev, RVScroll, PtRVData, ppPrnabl, CRVFData, RVTable,
  RVUni, RVItem, ppRichTxDrwCmd,
  ppRichViewReportDesign,
  {$IFDEF RV_USE_LLPDFLIB}
  RB_E_PDF,
  {$ENDIF}
  {$IFDEF REPORTBUILDER15}
  ppPDFRendererRV,
  {$ENDIF}
  Math;

procedure RVSetHelperDefaults(helper: TRVReportHelper);
begin
  with Helper.RichView do
  begin
    RVFTextStylesReadMode := rvf_sInsertMerge;
    RVFParaStylesReadMode := rvf_sInsertMerge;
    RVFOptions := RVFOptions + [rvfoSaveTextStyles, rvfoSaveParaStyles,
      rvfoSaveBack, rvfoLoadBack, rvfoSaveLayout, rvfoLoadLayout];
    RTFOptions := RTFOptions + [rvrtfPNGInsteadOfBitmap, rvrtfDuplicateUnicode];
    Options := Options + [rvoTagsArePChars];
    RTFReadProperties.TextStyleMode := rvrsAddIfNeeded;
    RTFReadProperties.ParaStyleMode := rvrsAddIfNeeded;
  end;
end;

{================================= TReportRVDataRef ===========================}
{
 TReportRVDataRef - TRVReportHelper with output canvas resolution = TPPRICHVIEWREFCANVASPPI
}
type
  TReportRVDataRef = class(TReportRVData)
  protected
    procedure GetSADForFormatting(Canvas: TCanvas;
      var sad: TRVScreenAndDevice); override;
  public
    procedure DataWriter(Stream: TStream); reintroduce;
    procedure DataReader(Stream: TStream);
  end;

procedure TReportRVDataRef.GetSADForFormatting(Canvas: TCanvas;
  var sad: TRVScreenAndDevice);
var
  ppi: Integer;
begin
  inherited;
  ppi := TReportRichViewRef(RichView).FInternalPixelsPerInch;
  if ppi > 0 then
  begin
    PrnSad.ppixDevice := ppi;
    PrnSad.ppiyDevice := ppi;
    sad.ppixDevice := ppi;
    sad.ppiyDevice := ppi;
  end;
end;

procedure TReportRVDataRef.DataWriter(Stream: TStream);
var
  StartPos, Size: Integer;
  Layout:         TRVLayoutInfo;
begin
  Size := 0;
  StartPos := Stream.Position;
  Stream.WriteBuffer(Size, SizeOf(Size));
  Layout := TCustomRichView(RichView).CreateLayoutInfo;
  GetRVData.SaveRVFToStream(Stream, False, TCustomRichView(RichView).Color,
    TCustomRichView(RichView).Background, Layout);
  Layout.Free;
  Size := Stream.Position - SizeOf(Size) - StartPos;
  Stream.Position := StartPos;
  Stream.WriteBuffer(Size, SizeOf(Size));
  Stream.Position := StartPos + SizeOf(Size) + Size;
end;

procedure TReportRVDataRef.DataReader(Stream: TStream);
var
  Size:      Integer;
  MemStream: TMemoryStream;
  Color:     TColor;
  Layout:    TRVLayoutInfo;
begin
  MemStream := TMemoryStream.Create;
  Layout := TRVLayoutInfo.Create;
  try
    Stream.ReadBuffer(Size, SizeOf(Size));
    MemStream.SetSize(Size);
    Stream.ReadBuffer(MemStream.Memory^, Size);
    Color := TCustomRichView(RichView).Color;
    LoadRVFFromStream(MemStream, Color, TCustomRichView(RichView).Background,
      Layout {$IFNDEF RVDONOTUSESTYLETEMPLATES}, nil, nil{$ENDIF});
    TCustomRichView(RichView).Color := Color;
    if Layout.Loaded then
      TCustomRichView(RichView).ApplyLayoutInfo(Layout);
  finally
    Layout.Free;
    MemStream.Free;
  end;
end;

constructor TReportRichViewRef.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FInternalPixelsPerInch := TPPRICHVIEWREFCANVASPPI;
  // Options := Options - [rvoFastFormatting];
end;

function TReportRichViewRef.GetDataClass: TRichViewRVDataClass;
begin
  Result := TReportRVDataRef;
end;

procedure TRVReportHelperRef.Clear;
begin
  FFormattedPageCount := 0;
  inherited;
end;

function TRVReportHelperRef.FormatNextPage(AMaxHeight: Integer): Boolean;
begin
  Result := not Finished;
  if Result then
  begin
    Result := inherited FormatNextPage(AMaxHeight);
    inc(FFormattedPageCount);
  end;
end;

procedure TRVReportHelperRef.Init(ACanvas: TCanvas; APageWidth: Integer);
begin
  FFormattedPageCount := 0;
  inherited;
end;

{$IFNDEF RVDONOTUSESEQ}

function TRVReportHelperRef.IgnoreEndnotes: Boolean;
begin
  Result := True;
end;
{$ENDIF}

procedure TRVReportHelperRef.Reset;
begin
  FFormattedPageCount := 0;
  inherited;
end;

function TRVReportHelperRef.CreateRichView: TCustomPrintableRV;
begin
  Result := TReportRichViewRef.Create(Self);
end;

{===================== TppCustomRichView (TppStretchable) =====================}
constructor TppCustomRichView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FCanvasRef := TCanvas.Create;
  FCanvasRef.Handle := CreateCompatibleDC(0);
  FCanvasRef.Font.PixelsPerInch := TPPRICHVIEWREFCANVASPPI;

  {init inherited properties}
  AutoSize := False;
  WordWrap := True;
  spHeight := 89;
  spWidth := 185;

  CacheRecordSize := SizeOf(TppCustomRichViewSaveRec);
  DrawCommandClass := TppDrawRichView;

  FHelper := TRVReportHelperRef.Create(nil);
  TReportRichViewRef(FHelper.RichView).FInternalPixelsPerInch :=
    TPPRICHVIEWREFCANVASPPI;

  FRVStyle := TRVStyle.Create(nil);
  FRVStyle.TextStyles.PixelsPerInch := TPPRICHVIEWREFCANVASPPI;
  FHelper.RichView.Style := FRVStyle;
  FHelper.RichView.HandleNeeded;

  RVSetHelperDefaults(FHelper);

  FCurrentPageNo := -1;
  FCurrentPageHeight := -1;
  FFirstPageHeight := -1;

  FInternalPixelsPerInch := TPPRICHVIEWREFCANVASPPI;
  FUsePrinterResolution := TPPRICHVIEWUSEPRINTERRES;
end;

{------------------------------------------------------------------------------}
destructor TppCustomRichView.Destroy;
var
  Handle: THandle;
begin
  try
    FHelper.Free;
  except
  end;
  Handle := FCanvasRef.Handle;
  FCanvasRef.Free;
  DeleteDC(Handle);
  FRVStyle.Free;
  inherited Destroy;
end;

{------------------------------------------------------------------------------}
function TppCustomRichView.GetRichView: TCustomRichView;
begin
  Result := FHelper.RichView;
end;

{------------------------------------------------------------------------------}
function TppCustomRichView.GetHelper: TRVReportHelper;
begin
  Result := FHelper;
end;

{------------------------------------------------------------------------------}
procedure TppCustomRichView.SetInternalPixelsPerInch(const Value: Integer);
begin
  FInternalPixelsPerInch := Value;
  if (FHelper <> nil) and not FUsePrinterResolution then
    TReportRichViewRef(FHelper.RichView).FInternalPixelsPerInch := Value;
end;

{------------------------------------------------------------------------------}
procedure TppCustomRichView.BoundsChange;
begin
  if not(pppcDesigning in DesignState) or (Printing) then
    Exit;
  // if (Band = nil) or (DesignControl = nil) then Exit;
  // if ((csReading in ComponentState) or (csLoading in ComponentState)) and (DesignControl = nil) then Exit;
  inherited BoundsChange;
end;

{------------------------------------------------------------------------------}
function TppCustomRichView.HasFont: Boolean;
begin
  Result := False;
end;

{------------------------------------------------------------------------------}
procedure TppCustomRichView.RestoreFromCache(aCachePageNo: Longint);
var
  lRec: TppCustomRichViewSaveRec;
begin
  ReadRecordFromCache(aCachePageNo, lRec);
  FRequiresReformatting := (FFirstPageHeight <> lRec.FFirstPageHeight);
  DrawComponent := lRec.FDrawComponent;
  DrawShifter := lRec.FDrawShifter;
  ObjectIndex := lRec.FObjectIndex;
  OverFlow := lRec.FOverFlow;
  KeptTogether := lRec.FKeptTogether;
  PartialGeneration := lRec.FPartialGeneration;
  FCurrentPageNo := lRec.FPageNo;
  FCurrentPageHeight := lRec.FPageHeight;
  FFirstPageHeight := lRec.FFirstPageHeight;
end;

{------------------------------------------------------------------------------}
procedure TppCustomRichView.SaveToCache(aCachePageNo: Longint);
var
  lRec: TppCustomRichViewSaveRec;

begin
  lRec.FDrawComponent := DrawComponent;
  lRec.FDrawShifter := DrawShifter;
  lRec.FObjectIndex := ObjectIndex;
  lRec.FOverFlow := OverFlow;
  lRec.FKeptTogether := KeptTogether;
  lRec.FPartialGeneration := PartialGeneration;
  lRec.FPageNo := FCurrentPageNo;
  lRec.FPageHeight := FCurrentPageHeight;
  lRec.FFirstPageHeight := FFirstPageHeight;

  WriteRecordToCache(aCachePageNo, lRec);
end;

{------------------------------------------------------------------------------}
procedure TppCustomRichView.UpdateDocument;
begin
end;

{------------------------------------------------------------------------------}
procedure TppCustomRichView.PropertiesToDrawCommand(aDrawCommand
  : TppDrawCommand);
begin
  inherited PropertiesToDrawCommand(aDrawCommand);
  if aDrawCommand is TppDrawRichView then
  begin
    if FUsePrinterResolution then
    begin
      TppDrawRichView(aDrawCommand).FInternalPixelsPerInch := 0;
      TppDrawRichView(aDrawCommand).FPrinterPixelsPerInch :=
        Report.Printer.Canvas.Font.PixelsPerInch;
      TppDrawRichView(aDrawCommand).FCanvasRef := Report.Printer.Canvas;
      TppDrawRichView(aDrawCommand).FFreeCanvas := False;
    end
    else
    begin
      TppDrawRichView(aDrawCommand).FInternalPixelsPerInch :=
        InternalPixelsPerInch;
      TppDrawRichView(aDrawCommand).FPrinterPixelsPerInch := 0;
      TppDrawRichView(aDrawCommand).FCanvasRef := TCanvas.Create;
      TppDrawRichView(aDrawCommand).FCanvasRef.Handle := CreateCompatibleDC(0);
      TppDrawRichView(aDrawCommand).FCanvasRef.Font.PixelsPerInch :=
        InternalPixelsPerInch;
      TppDrawRichView(aDrawCommand).FFreeCanvas := True;
    end;
    TppDrawRichView(aDrawCommand).Transparent := Transparent;
    TppDrawRichView(aDrawCommand).Color := Color;
    TppDrawRichView(aDrawCommand).Left := PrintPosRect.Left;
    TppDrawRichView(aDrawCommand).Top := PrintPosRect.Top;
    TppDrawRichView(aDrawCommand).Height := PrintPosRect.Bottom -
      PrintPosRect.Top;
    TppDrawRichView(aDrawCommand).Width := PrintPosRect.Right -
      PrintPosRect.Left;
    TppDrawRichView(aDrawCommand).FPageHeight := FCurrentPageHeight;
    TppDrawRichView(aDrawCommand).FStream.Free;
    if FCurrentPageNo >= 0 then
    begin
      TppDrawRichView(aDrawCommand).FStream := TMemoryStream.Create;
      FHelper.SavePageAsRVF(TppDrawRichView(aDrawCommand).FStream,
        FCurrentPageNo);
    end;
    TppDrawRichView(aDrawCommand).FPageWidth :=
      FHelper.RichView.RVData.TextWidth +
      RV_XToDevice(FHelper.RichView.LeftMargin + FHelper.RichView.RightMargin,
      TReportRVDataRef(FHelper.RichView.RVData).PrnSad);
  end;
end;

{------------------------------------------------------------------------------}
procedure TppCustomRichView.CalcSpaceUsed;
var
  i, H:        Integer;
  Initialized: Boolean;
begin
  UpdateDocument;
  Initialized := False;

  FCurrentPageHeight :=
    Trunc(ppFromMMThousandths(PrintPosRect.Bottom - PrintPosRect.Top, utInches,
    pprtVertical, nil) * GetLPY);

  if not OverFlow then
  begin
    FCurrentPageNo := -1;
    KeptTogether := False;
  end;

  if FRequiresReformatting or (not Helper.Finished and (FCurrentPageNo >= 1) and
    (FCurrentPageNo >= TRVReportHelperRef(FHelper).FFormattedPageCount)) then
  begin
    if not FRequiresReformatting then
      Helper.Reset
    else
      InitHelper;
    Initialized := True;
    if FCurrentPageNo >= 1 then
      FHelper.FormatNextPage(FFirstPageHeight);
    for i := 2 to FCurrentPageNo do
    begin
      if FHelper.Finished then
      begin
        FCurrentPageNo := -1;
        break;
      end;
      FHelper.FormatNextPage(FCurrentPageHeight);
    end;
    FRequiresReformatting := False;
  end;

  if (FCurrentPageNo < 1) or
    (FHelper.Finished and (FCurrentPageNo >= FHelper.PagesCount)) then
  begin
    if Initialized then
      Helper.Reset
    else
      InitHelper;
    FCurrentPageNo := 1;
  end
  else
    inc(FCurrentPageNo);

  H := FCurrentPageHeight;
  if (ParentStretch and Stretch) and (FCurrentPageNo = 1) and
    not KeptTogether and ((MINPPRICHVIEWHEIGHTMM <= 0) or
    ((PrintPosRect.Bottom - PrintPosRect.Top) < MINPPRICHVIEWHEIGHTMM * 1000))
    and not(Helper.RichView.RVData as TCustomMultiPagePtblRVData)
    .CanPlaceFirstPageHere(H, False, TReportRVDataRef(FHelper.RichView.RVData)
    .PrnSad, False, False) then
  begin
    OverFlow := True;
    SpaceUsed := 1;
    FCurrentPageNo := -1;
    KeptTogether := True;
    Exit;
  end;

  if not FHelper.Finished then
    FHelper.FormatNextPage(FCurrentPageHeight);
  if FHelper.Finished and (FHelper.PagesCount = 0) then
    FCurrentPageHeight := 0
  else if FHelper.Finished and (FCurrentPageNo = FHelper.PagesCount) then
  begin
    if FHelper.GetLastPageHeight < FCurrentPageHeight then
      FCurrentPageHeight := FHelper.GetLastPageHeight;
  end;
  if ParentStretch and Stretch then
  begin
    OverFlow := not FHelper.Finished;
    if OverFlow and ShouldKeepTogether then
    begin
      SpaceUsed := 1;
      FCurrentPageNo := -1;
      KeptTogether := True;
      Exit;
    end;
    SpaceUsed := Trunc(ppToMMThousandths(FCurrentPageHeight, utInches,
      pprtVertical, nil) / GetLPY)
  end
  else
  begin
    OverFlow := False;
    SpaceUsed := mmHeight;
    FCurrentPageHeight := Trunc(ppFromMMThousandths(mmHeight, utInches,
      pprtVertical, nil) * GetLPY);
  end;
  if FCurrentPageNo = 1 then
    FFirstPageHeight := FCurrentPageHeight;
end;

{------------------------------------------------------------------------------}
procedure TppCustomRichView.Clear;
begin
  FHelper.Clear;
  FHelper.Reset;
end;

{------------------------------------------------------------------------------}
procedure TppCustomRichView.Init;
begin
  inherited;
  FCurrentPageNo := -1;
  FCurrentPageHeight := -1;
  FFirstPageHeight := -1;
  FRequiresReformatting := False;
end;

{------------------------------------------------------------------------------}
procedure TppCustomRichView.InitHelper;
begin
  if UsePrinterResolution then
  begin
    TReportRichViewRef(FHelper.RichView).FInternalPixelsPerInch := 0;
    FHelper.RichView.Style.TextStyles.PixelsPerInch := 0;
    FHelper.Init(Report.Printer.Canvas,
      Trunc(ppFromMMThousandths(mmWidth, utInches, pprtHorizontal,
      nil) * GetLPX));
  end
  else
  begin
    FCanvasRef.Font.PixelsPerInch := FInternalPixelsPerInch;
    FHelper.RichView.Style.TextStyles.PixelsPerInch := FInternalPixelsPerInch;
    FHelper.Init(FCanvasRef, Trunc(ppFromMMThousandths(mmWidth, utInches,
      pprtHorizontal, nil) * GetLPX));
  end;
end;

{------------------------------------------------------------------------------}
procedure ConvertRVToUnicode(RVData: TCustomRVData);
var
  i, r, c, StyleNo: Integer;
  table:            TRVTableItemInfo;
begin
  for i := 0 to RVData.ItemCount - 1 do
  begin
    StyleNo := RVData.GetItemStyle(i);
    if StyleNo >= 0 then
    begin
      if not RVData.GetRVStyle.TextStyles[StyleNo].Unicode then
      begin
        RVData.SetItemTextR(i, RVU_GetRawUnicode(RVData.GetItemTextW(i)));
        Include(RVData.GetItem(i).ItemOptions, rvioUnicode);
      end;
    end
    else if RVData.GetItemStyle(i) = rvsTable then
    begin
      table := TRVTableItemInfo(RVData.GetItem(i));
      for r := 0 to table.Rows.Count - 1 do
        for c := 0 to table.Rows[r].Count - 1 do
          if table.Cells[r, c] <> nil then
            ConvertRVToUnicode(table.Cells[r, c].GetRVData);
    end;
  end;
end;

procedure ConvertRVFromUnicode(RVData: TCustomRVData);
var
  i, r, c, StyleNo: Integer;
  table:            TRVTableItemInfo;
begin
  for i := 0 to RVData.ItemCount - 1 do
  begin
    StyleNo := RVData.GetItemStyle(i);
    if StyleNo >= 0 then
    begin
      if RVData.GetRVStyle.TextStyles[StyleNo].Unicode then
      begin
        RVData.Items[i] := RVData.GetItemTextA(i);
        Exclude(RVData.GetItem(i).ItemOptions, rvioUnicode);
      end;
    end
    else if RVData.GetItemStyle(i) = rvsTable then
    begin
      table := TRVTableItemInfo(RVData.GetItem(i));
      for r := 0 to table.Rows.Count - 1 do
        for c := 0 to table.Rows[r].Count - 1 do
          if table.Cells[r, c] <> nil then
            ConvertRVFromUnicode(table.Cells[r, c].GetRVData);
    end;
  end;
end;

procedure ConvertToUnicode(rv: TCustomRichView);
var
  i: Integer;
begin
  ConvertRVToUnicode(rv.RVData);
  for i := 0 to rv.Style.TextStyles.Count - 1 do
    rv.Style.TextStyles[i].Unicode := True;
end;

procedure ConvertFromUnicode(rv: TCustomRichView);
var
  i: Integer;
begin
  ConvertRVFromUnicode(rv.RVData);
  for i := 0 to rv.Style.TextStyles.Count - 1 do
    rv.Style.TextStyles[i].Unicode := False;
end;

{------------------------------------------------------------------------------}
procedure TppCustomRichView.SetUnicode(const Value: Boolean);
begin
  FUnicode := Value;
  if Unicode then
    RichView.RTFReadProperties.UnicodeMode := rvruOnlyUnicode
  else
    RichView.RTFReadProperties.UnicodeMode := rvruNoUnicode;
  if Unicode then
    ConvertToUnicode(RichView)
  else
    ConvertFromUnicode(RichView);
end;

{------------------------------------------------------------------------------}
function TppCustomRichView.GetLPX: Integer;
begin
  if FUsePrinterResolution then
    Result := GetDeviceCaps(Report.Printer.Canvas.Handle, LOGPIXELSX)
  else
    Result := FInternalPixelsPerInch;
end;

{------------------------------------------------------------------------------}
function TppCustomRichView.GetLPY: Integer;
begin
  if FUsePrinterResolution then
    Result := GetDeviceCaps(Report.Printer.Canvas.Handle, LOGPIXELSY)
  else
    Result := FInternalPixelsPerInch;
end;

{================= TppRichView (TppCustomRichView) ============================}
constructor TppRichView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  DefaultPropName := 'Caption';
  DefaultPropEditType := etAutoEdit;

end;

{------------------------------------------------------------------------------}
procedure TppRichView.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineBinaryProperty('Document', TReportRVDataRef(RichView.RVData)
    .DataReader, TReportRVDataRef(RichView.RVData).DataWriter,
    RichView.ItemCount > 0);
end;

{=========================== TppDBRichView (TppCustomRichView) ================}
constructor TppDBRichView.Create(AOwner: TComponent);
begin
  inherited;
  FAutoDisplay := True;
  FCurRecordNo := -1;
end;

{------------------------------------------------------------------------------}
procedure TppDBRichView.CalcSpaceUsed;
begin
  inherited;
end;

{------------------------------------------------------------------------------}
function TppDBRichView.IsDataAware: Boolean;
begin
  Result := True;
end;

procedure TppDBRichView.Init;
begin
  inherited;
  FCurRecordNo := -1;
end;

{------------------------------------------------------------------------------}
procedure TppDBRichView.Load(Stream: TStream; Unicode: Boolean);
var
  s: TRVRawByteString;
const
  RTF_START = '{\rtf';
  {...................................................}
  function IsRTFW(Stream: TStream): Boolean;
  var
    DataStart: TRVUnicodeString;
  begin
    Result := (Stream.Size > 10);
    if not Result then
      Exit;
    SetLength(DataStart, 5);
    Stream.ReadBuffer(Pointer(DataStart)^, 10);
    Result := DataStart = RTF_START;
    Stream.Position := 0;
  end;
{...................................................}
  function IsRTFA(Stream: TStream): Boolean;
  var
    DataStart: TRVAnsiString;
  begin
    Result := (Stream.Size > 5);
    if not Result then
      Exit;
    SetLength(DataStart, 5);
    Stream.ReadBuffer(Pointer(DataStart)^, 5);
    Result := DataStart = RTF_START;
    Stream.Position := 0;
  end;
{...................................................}
  function IsRTF(Stream: TStream): Boolean;
  begin
    if Unicode then
      Result := IsRTFW(Stream)
    else
      Result := IsRTFA(Stream);
  end;
{...................................................}
  function AllZero(const s: TRVRawByteString): Boolean;
  var
    i: Integer;
  begin
    Result := False;
    for i := 1 to Length(s) do
      if s[i] <> #0 then
        Exit;
    Result := True;
  end;
{...................................................}
{$IFDEF RV_USE_HTML}
  function LoadHTML(const s: TRVRawByteString): Boolean;
  var
    html: TRVHTMLImporter;
    s2:   TRVRawByteString;
  begin
    Result := True;
    try
      if ppExternalHtmlImporter = nil then
      begin
        html := TRVHTMLImporter.Create(nil);
        try
          if Unicode then
          begin
            // s contains raw Unicode; converting for loading as UTF-8
            html.Encoding := rvhtmleUTF8;
            s2 := UTF8Encode(RVU_RawUnicodeToWideString(s));
          end
          else
            // s contains either ANSI or UTF-8; loading as it is
            s2 := s;
          html.RichView := RichView;
          html.LoadHTML(s2);
        finally
          html.Free;
        end;
      end
      else
      begin
        if Unicode then
        begin
          // s contains raw Unicode; converting to UTF-8 with BOM
          // BOM forces to read this HTML as UTF-8, even if the importer expects ANSI
          s2 := UTF8Encode(RVU_RawUnicodeToWideString(s));
          if (Length(s2) < 3) or (s2[1] <> #$EF) or (s2[2] <> #$BB) or
            (s2[3] <> #$BF) then
            s2 := #$EF#$BB#$BF + s2;
        end
        else
        begin
          // s contains either ANSI or UTF-8; loading as it is
          s2 := s;
        end;
        ppExternalHtmlImporter.RichView := RichView;
        ppExternalHtmlImporter.LoadHTML(s2);
      end;
    except
      Result := False;
    end;
  end;
{$ENDIF}
{...................................................}
  function LoadRTF(Stream: TStream): Boolean;
  var
    LStream: TMemoryStream;
    sW:      TRVUnicodeString;
    sA:      TRVAnsiString;
  begin
    if Unicode then
    begin
      SetLength(sW, Stream.Size div 2);
      Stream.ReadBuffer(Pointer(sW)^, Stream.Size);
      sA := TRVAnsiString(sW);
      LStream := TMemoryStream.Create;
      try
        LStream.WriteBuffer(PRVAnsiChar(sA)^, Length(sA));
        LStream.Position := 0;
        Result := RichView.LoadRTFFromStream(LStream);
      finally
        LStream.Free;
      end;
    end
    else
      Result := RichView.LoadRTFFromStream(Stream);
  end;

{...................................................}
begin
  {$IFDEF RVALOWEMPTYDBRICHVIEW}
  if Stream.Size = 0 then
  begin
    Clear;
    exit;
  end;
  {$ENDIF}
  Stream.Position := 0;
  if Unicode or not RichView.LoadRVFFromStream(Stream) then
  begin
    Stream.Position := 0;
    RichView.Clear;
    if not(IsRTF(Stream) and LoadRTF(Stream)) then
    begin
      Stream.Position := 0;
      SetLength(s, Stream.Size);
      Stream.ReadBuffer(PRVAnsiChar(s)^, Stream.Size);
      RichView.Clear;
      if AllZero(s) then
        s := '';
      {$IFDEF RV_USE_HTML}
      if not LoadHTML(s) then
      {$ENDIF}
      begin
        RichView.Clear;
        if Unicode then
          RichView.AddTextNLW(RVU_RawUnicodeToWideString(s), 0, 0, 0, False)
        else
          RichView.AddTextNLA(s, 0, 0, 0);
      end;
    end;
  end;
  if Assigned(OnLoadDocument) then
    OnLoadDocument(Self);
  InvalidateDesignControl;
end;

{------------------------------------------------------------------------------}
procedure TppDBRichView.LoadFromField(Updating: Boolean);
var
  Stream:  TRVMemoryStream;
  Unicode: Boolean;
  sA:      TRVAnsiString;
  s:       String;
begin
  if not(CheckForPipelineData) then
    Exit;

  if not(FAutoDisplay) then
  begin
    Clear;
    InvalidateDesignControl;
    Exit;
  end;
  if not Updating or (DataPipeline.RecordNo <> FCurRecordNo) then
  begin
    Stream := TRVMemoryStream.Create;
    try
      {$IFDEF RVUNICODESTR}
      Unicode := DataPipeline.GetFieldDataType(DataField) <> dtBlob;
      {$ELSE}
      Unicode := False;
      {$ENDIF}
      DataPipeline.GetFieldAsStream(DataField, Stream);
      Stream.Position := 0;
      Load(Stream, Unicode);
      if Assigned(FOnGetRichText) then
      begin
        Stream.Clear;
        RichView.SaveRTFToStream(Stream, False);
        SetLength(sA, Stream.Size);
        Stream.Position := 0;
        Stream.ReadBuffer(PRVAnsiChar(sA)^, Length(sA));
        s := String(sA);
        FOnGetRichText(Self, s);
        sA := TRVAnsiString(s);
        Stream.Clear;
        Stream.WriteBuffer(PRVAnsiChar(sA)^, Length(sA));
        RichView.Clear;
        RichView.LoadRTFFromStream(Stream);
      end;
      FCurRecordNo := DataPipeline.RecordNo;
      FRequiresReformatting := True;
    finally
      Stream.Free;
    end;
  end;
end;

{------------------------------------------------------------------------------}
procedure TppDBRichView.Notify(aCommunicator: TppCommunicator;
  aOperation: TppOperationType);
begin
  inherited Notify(aCommunicator, aOperation);
  if (csDestroying in ComponentState) then
    Exit;
  if (aOperation in [ppopDataChange, ppopRemove]) and
    (pppcDesigning in DesignState) and not(Printing) then
    LoadFromField(False);
end;

{------------------------------------------------------------------------------}
procedure TppDBRichView.SetAutoDisplay(aValue: Boolean);
begin
  inherited;
  LoadFromField(False);
end;

{------------------------------------------------------------------------------}
procedure TppDBRichView.SetDataField(const Value: String);
begin
  inherited;
  LoadFromField(False);
end;

{------------------------------------------------------------------------------}
procedure TppDBRichView.UpdateDocument;
begin
  LoadFromField(True);
  // FCurrentPageNo := -1;
  // FCurrentPageHeight := -1;
end;

{======================== TppDrawRichView (TppDrawCommand) ====================}
constructor TppDrawRichView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  DataType := dtBlob;
  FTransparent := False;

end;

{------------------------------------------------------------------------------}
destructor TppDrawRichView.Destroy;
var
  Handle: THandle;
begin
  FStream.Free;
  FStream := nil;

  if FFreeCanvas then
  begin
    Handle := FCanvasRef.Handle;
    FCanvasRef.Free;
    DeleteDC(Handle);
  end;

  inherited Destroy;
end;

{------------------------------------------------------------------------------}
procedure TppDrawRichView.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineBinaryProperty('Stream', StreamReader, StreamWriter,
    FStream <> nil);
end;

{------------------------------------------------------------------------------}
procedure TppDrawRichView.StreamReader(Stream: TStream);
var
  v: Integer;
begin
  Stream.ReadBuffer(v, SizeOf(Integer));
  FStream.Free;
  FStream := TMemoryStream.Create;
  FStream.SetSize(v);
  Stream.ReadBuffer(FStream.Memory^, v);
end;

{------------------------------------------------------------------------------}
procedure TppDrawRichView.StreamWriter(Stream: TStream);
var
  v: Integer;
begin
  v := FStream.Size;
  Stream.WriteBuffer(v, SizeOf(Integer));
  Stream.WriteBuffer(FStream.Memory^, v);
end;
{------------------------------------------------------------------------------}
{$IFDEF REPORTBUILDER15}

function TppDrawRichView.AsMetaFile: TMetafile;
begin
  Result := GetAsMetafile(False);
end;
{$ENDIF}

{------------------------------------------------------------------------------}
procedure TppDrawRichView.Assign(Source: TPersistent);
begin
  inherited Assign(Source);

  if (Source is TppDrawRichView) then
  begin
    if FFreeCanvas then
      FCanvasRef.Free;
    if TppDrawRichView(Source).FFreeCanvas then
    begin
      FCanvasRef := TCanvas.Create;
      FCanvasRef.Handle := CreateCompatibleDC(0);
      FCanvasRef.Font.PixelsPerInch := TppDrawRichView(Source)
        .FCanvasRef.Font.PixelsPerInch;
      FPrinterPixelsPerInch := 0;
    end
    else
    begin
      FCanvasRef := TppDrawRichView(Source).FCanvasRef;
      if FCanvasRef <> nil then
        FPrinterPixelsPerInch := FCanvasRef.Font.PixelsPerInch
      else
        FPrinterPixelsPerInch := TppDrawRichView(Source).FPrinterPixelsPerInch;
    end;
    FFreeCanvas := TppDrawRichView(Source).FFreeCanvas;
    FColor := TppDrawRichView(Source).Color;
    FTransparent := TppDrawRichView(Source).Transparent;
    FPageHeight := TppDrawRichView(Source).FPageHeight;
    FPageWidth := TppDrawRichView(Source).FPageWidth;
    FInternalPixelsPerInch := TppDrawRichView(Source).FInternalPixelsPerInch;
    FStream.Free;
    if TppDrawRichView(Source).FStream <> nil then
    begin
      FStream := TMemoryStream.Create;
      TppDrawRichView(Source).FStream.SaveToStream(FStream);
    end
    else
      FStream := nil;
  end;
end;
{------------------------------------------------------------------------------}
{$IFDEF RV_USE_PRAGNAAN}

type
  THackPsRBDevice = class(TPsRBExportDevice);
  {$ENDIF}
  {$IFDEF REPORTBUILDER12}

type
  TppScrGrDevide = TppGraphicsDevice;
  {$ELSE}

type
  TppScrGrDevide = TppScreenDevice;
  {$ENDIF}

function TppDrawRichView.DoDraw(FHelper: TRVReportHelper;
  aDevice: TppDevice): Boolean;
var
  lCanvas:    TCanvas;
  ppi:        Integer;
  SetLeftTop: Boolean;
  ADrawRect:  TRect;
  FreeCanvas: Boolean;
  {$IFDEF RV_USE_PRAGNAAN}
  gr:         TGraphic;
  ADrawImage: TppDrawImage;
  ADrawRTF:   TppDrawRichText;
  {$ENDIF}
begin
  Result := True;
  if FStream = nil then
    Exit;
  if Width <= 0 then
    Width := 1;
  if Height <= 0 then
    Height := 1;
  SetLeftTop := True;
  FreeCanvas := False;
  {$IFDEF RV_USE_PRAGNAAN}
  ADrawImage := nil;
  gr := nil;
  {$ENDIF}
  if (aDevice is TppPrinterDevice) then
  begin
    TppPrinterDevice(aDevice).CalcDrawPosition(Self);
    lCanvas := TppPrinterDevice(aDevice).Printer.Canvas;
    ADrawRect := DrawRect;
  end
  else if (aDevice is TppScrGrDevide) then
  begin
    TppScrGrDevide(aDevice).ScaleVertices(Self);
    lCanvas := TppScrGrDevide(aDevice).Canvas;
    if ppPrinter <> nil then
      ppPrinter.PrinterSetup := Page.PrinterSetup;
    ADrawRect := DrawRect;
  end
  {$IFDEF RV_USE_PRAGNAAN}
  else if ((aDevice is TPsRBRTFDevice) and not RVppsInRTFAsImages)
  {$IFDEF RV_PRAGNAAN_PRO}
    or ((aDevice is TPsRBExcelDevice) and
    (PsRBExcelSettings.RichTextEncodingType = rtPlainText))
  {$ELSE}
    or (aDevice is TPsRBExcelDevice)
  {$ENDIF}
  then
  begin
    ADrawRTF := TppDrawRichText.Create(nil);
    ADrawRTF.Left := Left;
    ADrawRTF.Top := Top;
    ADrawRTF.Width := Width;
    ADrawRTF.Height := Height;
    ADrawRTF.Color := Color;
    ADrawRTF.Transparent := Transparent;
    FHelper.RichView.RTFOptions := [rvrtfDuplicateUnicode, rvrtfSaveEMFAsWMF,
      rvrtfSaveBitmapDefault];
    FHelper.RichView.SaveRTFToStream(ADrawRTF.RichTextStream, False);
    ADrawRTF.StartCharPos := 0;
    ADrawRTF.EndCharPos := FStream.Size;
    THackPsRBDevice(aDevice).DrawRichText(ADrawRTF);
    ADrawRTF.Free;
    Exit;
  end
  else if (aDevice is TPsRBExportDevice) then
  begin
    ADrawImage := TppDrawImage.Create(nil);
    ADrawImage.Left := Left;
    ADrawImage.Top := Top;
    ADrawImage.Width := Width;
    ADrawImage.Height := Height;
    if (aDevice is TPsRBPDFDevice) or (aDevice is TPsRBRTFDevice) or
      (aDevice is TPsRBExcelDevice) then
    begin
      gr := TMetafile.Create;
      lCanvas := TMetafileCanvas.Create(TMetafile(gr), 0);
      FreeCanvas := True;
    end
    else
    begin
      gr := TBitmap.Create;
      lCanvas := TBitmap(gr).Canvas;
    end;
    {$IFDEF RV_PRAGNAAN_PRO}
    gr.Width := Round(MMThousandthsToPixel(Width));
    gr.Height := Round(MMThousandthsToPixel(Height));
    {$ELSE}
    gr.Width := Round(THackPsRBDevice(aDevice).ThousandthOfMMToPixel(Width));
    gr.Height := Round(THackPsRBDevice(aDevice).ThousandthOfMMToPixel(Height));
    {$ENDIF}
    SetLeftTop := False;
    ADrawRect := Rect(0, 0, gr.Width, gr.Height);
  end
  {$ENDIF}
  {$IFDEF RV_USE_LLPDFLIB}
  else if (aDevice is TllPDFDevice) then
  begin
    lCanvas := TllPDFDevice(aDevice).Canvas;
    ADrawRect := DrawRect;
  end
  {$ENDIF}
  else
  begin
    lCanvas := nil;
    Result := False;
  end;

  if Result then
  begin
    if Transparent then
      lCanvas.CopyMode := cmSrcAnd
    else
    begin
      lCanvas.CopyMode := cmSrcCopy;
      lCanvas.Brush.Style := bsSolid;
      lCanvas.Brush.Color := FColor;
      lCanvas.FillRect(ADrawRect);
    end;
    lCanvas.Brush.Style := bsClear;

    with ADrawRect do
      IntersectClipRect(lCanvas.Handle, Left, Top, Right, Bottom);

    SetMapMode(lCanvas.Handle, MM_ANISOTROPIC);
    SetWindowExtEx(lCanvas.Handle, FPageWidth, FPageHeight, nil);
    ppi := lCanvas.Font.PixelsPerInch;
    lCanvas.Font.PixelsPerInch := FCanvasRef.Font.PixelsPerInch;
    with ADrawRect do
    begin
      SetViewportExtEx(lCanvas.Handle, Right - Left, Bottom - Top, nil);
      if SetLeftTop then
        SetViewportOrgEx(lCanvas.Handle, Left, Top, nil);
    end;
    FHelper.TransparentBackground := FTransparent;
    if not Transparent then
      FHelper.RichView.Color := Color;
    if (aDevice is TppPrinterDevice) then
      FHelper.DrawPage(1, lCanvas, False, FPageHeight)
    else if (aDevice is TppScrGrDevide)
    {$IFDEF RV_USE_LLPDFLIB} or (aDevice is TllPDFDevice) {$ENDIF}
    {$IFDEF RV_USE_PRAGNAAN} or (aDevice is TPsRBExportDevice) {$ENDIF}
    then
      FHelper.DrawPage(1, lCanvas, True, FPageHeight);
    lCanvas.Font.PixelsPerInch := ppi;
    SetMapMode(lCanvas.Handle, MM_TEXT);
    SetViewportOrgEx(lCanvas.Handle, 0, 0, nil);
    SelectClipRgn(lCanvas.Handle, 0);
    if FreeCanvas then
      lCanvas.Free;
  end;
  {$IFDEF RV_USE_PRAGNAAN}
  if ADrawImage <> nil then
  begin
    ADrawImage.Picture.Graphic := gr;
    gr.Free;
    THackPsRBDevice(aDevice).DrawImage(ADrawImage);
    ADrawImage.Free;
  end;
  {$ENDIF}
end;

{------------------------------------------------------------------------------}
procedure TppDrawRichView.ValidateCanvasRef;
begin
  if FCanvasRef <> nil then
    Exit;
  if (FInternalPixelsPerInch = 0) and (FPrinterPixelsPerInch = 0) then
    FInternalPixelsPerInch := TPPRICHVIEWREFCANVASPPI;

  if FInternalPixelsPerInch = 0 then
  begin
    // FCanvasRef  := Report.Printer.Canvas;
    // FFreeCanvas := False;
    FCanvasRef := TCanvas.Create;
    FCanvasRef.Handle := CreateCompatibleDC(0);
    FCanvasRef.Font.PixelsPerInch := FPrinterPixelsPerInch;
    FInternalPixelsPerInch := FPrinterPixelsPerInch;
    FPrinterPixelsPerInch := 0;
    FFreeCanvas := True;
  end
  else
  begin
    FCanvasRef := TCanvas.Create;
    FCanvasRef.Handle := CreateCompatibleDC(0);
    FCanvasRef.Font.PixelsPerInch := FInternalPixelsPerInch;
    FFreeCanvas := True;
  end;
end;

{------------------------------------------------------------------------------}
function TppDrawRichView.Draw(aDevice: TppDevice): Boolean;
var
  lHelper:  TRVReportHelper;
  lRVStyle: TRVStyle;
begin
  if FStream = nil then
  begin
    Result := True;
    Exit;
  end;
  ValidateCanvasRef;
  lHelper := TRVReportHelperRef.Create(nil);
  lRVStyle := TRVStyle.Create(nil);
  try
    TReportRichViewRef(lHelper.RichView).FInternalPixelsPerInch :=
      FInternalPixelsPerInch;
    if FInternalPixelsPerInch > 0 then
      FCanvasRef.Font.PixelsPerInch := FInternalPixelsPerInch;
    lRVStyle.TextStyles.PixelsPerInch := FInternalPixelsPerInch;
    RVSetHelperDefaults(lHelper);
    lHelper.RichView.Style := lRVStyle;
    FStream.Position := 0;
    lHelper.RichView.LoadRVFFromStream(FStream);
    lHelper.Init(FCanvasRef, FPageWidth);
    lHelper.FormatNextPage(FPageHeight + 100);
    Result := DoDraw(lHelper, aDevice);
  finally
    lHelper.Free;
    lRVStyle.Free;
  end;
end;

function TppDrawRichView.GetAsMetafile(AllowMetafilesInside: Boolean)
  : TMetafile;
var
  lCanvas: TMetafileCanvas;
  // lHelper: TRVReportHelper;
  // lRVStyle: TRVStyle;
  lWidth, lHeight: Integer;
  lPrinter:        TppPrinter;
begin
  lPrinter := ppPrinter;
  lWidth := Trunc(ppFromMMThousandths(Width, utPrinterPixels, pprtHorizontal,
    lPrinter));
  lHeight := Trunc(ppFromMMThousandths(Height, utPrinterPixels, pprtVertical,
    lPrinter));
  Result := TMetafile.Create;
  Result.Width := lWidth;
  Result.Height := lHeight;
  if FStream <> nil then
  begin
    ValidateCanvasRef;
    lCanvas := TMetafileCanvas.Create(Result, FCanvasRef.Handle);
    DrawTo(lCanvas, Rect(0, 0, lWidth, lHeight), True, AllowMetafilesInside);
    lCanvas.Free;
  end;
  {
   ValidateCanvasRef;
   Result := TMetafile.Create;
   Result.Width := FPageWidth;
   Result.Height := FPageHeight;
   lCanvas := TMetafileCanvas.Create(Result, FCanvasRef.Handle);
   lCanvas.Font.PixelsPerInch := FCanvasRef.Font.PixelsPerInch;
   lHelper := TRVReportHelperRef.Create(nil);
   lRVStyle := TRVStyle.Create(nil);
   try
   TReportRichViewRef(lHelper.RichView).FInternalPixelsPerInch := FInternalPixelsPerInch;
   if FInternalPixelsPerInch>0 then
   FCanvasRef.Font.PixelsPerInch := FInternalPixelsPerInch;
   lRVStyle.TextStyles.PixelsPerInch := FInternalPixelsPerInch;
   SetHelperDefaults(lHelper);
   lHelper.RichView.Style := lRVStyle;
   FStream.Position := 0;
   lHelper.RichView.LoadRVFFromStream(FStream);
   lHelper.Init(FCanvasRef, FPageWidth);
   lHelper.FormatNextPage(FPageHeight+100);
   lHelper.DrawPage(1, lCanvas, True, FPageHeight);
   finally
   lCanvas.Free;
   lHelper.Free;
   lRVStyle.Free;
   end;
  }
end;

procedure TppDrawRichView.DrawTo(lCanvas: TCanvas; ADrawRect: TRect;
  MetafileCompatible, AllowMetafiles: Boolean);
var
  lHelper:  TRVReportHelper;
  lRVStyle: TRVStyle;
  ppi:      Integer;
begin
  if FStream = nil then
    Exit;
  ValidateCanvasRef;
  lHelper := TRVReportHelperRef.Create(nil);
  lRVStyle := TRVStyle.Create(nil);

  try
    lHelper.NoMetafiles := not AllowMetafiles;
    lHelper.MetafileCompatibility := MetafileCompatible;
    lRVStyle.TextStyles.PixelsPerInch := FInternalPixelsPerInch;
    TReportRichViewRef(lHelper.RichView).FInternalPixelsPerInch :=
      FInternalPixelsPerInch;
    if FInternalPixelsPerInch > 0 then
      FCanvasRef.Font.PixelsPerInch := FInternalPixelsPerInch;
    RVSetHelperDefaults(lHelper);
    lHelper.RichView.Style := lRVStyle;
    FStream.Position := 0;
    lHelper.RichView.LoadRVFFromStream(FStream);
    lHelper.Init(FCanvasRef, FPageWidth);
    lHelper.FormatNextPage(FPageHeight + 100);

    if Transparent then
    begin
      lCanvas.CopyMode := cmSrcAnd;
    end
    else
    begin
      lCanvas.CopyMode := cmSrcCopy;
      lCanvas.Brush.Style := bsSolid;
      lCanvas.Brush.Color := FColor;
      lCanvas.FillRect(ADrawRect);
    end;

    lCanvas.Brush.Style := bsClear;

    SetMapMode(lCanvas.Handle, MM_ANISOTROPIC);
    SetWindowExtEx(lCanvas.Handle, FPageWidth, FPageHeight, Nil);

    ppi := lCanvas.Font.PixelsPerInch;
    if FInternalPixelsPerInch > 0 then
      lCanvas.Font.PixelsPerInch := FInternalPixelsPerInch;

    SetViewportExtEx(lCanvas.Handle, ADrawRect.Right - ADrawRect.Left,
      ADrawRect.Bottom - ADrawRect.Top, Nil);
    SetViewportOrgEx(lCanvas.Handle, ADrawRect.Left, ADrawRect.Top, Nil);

    IntersectClipRect(lCanvas.Handle, 0, 0, FPageWidth, FPageHeight);

    lHelper.TransparentBackground := FTransparent;
    if not Transparent then
      lHelper.RichView.Color := Color;

    lHelper.DrawPage(1, lCanvas, False, FPageHeight);

    lCanvas.Font.PixelsPerInch := ppi;

    SelectClipRgn(lCanvas.Handle, 0);

    SetMapMode(lCanvas.Handle, MM_TEXT);
    SetViewportOrgEx(lCanvas.Handle, 0, 0, Nil);

  finally
    lHelper.Free;
    lRVStyle.Free;
  end;
end;

procedure TppDrawRichView.SaveToStream(AStream: TStream);
var
  lHelper:  TRVReportHelper;
  lRVStyle: TRVStyle;
begin
  if FStream = nil then
    Exit;
  lHelper := TRVReportHelperRef.Create(nil);
  lRVStyle := TRVStyle.Create(nil);
  try
    lRVStyle.TextStyles.PixelsPerInch := FInternalPixelsPerInch;
    TReportRichViewRef(lHelper.RichView).FInternalPixelsPerInch :=
      FInternalPixelsPerInch;
    if FInternalPixelsPerInch > 0 then
      FCanvasRef.Font.PixelsPerInch := FInternalPixelsPerInch;
    RVSetHelperDefaults(lHelper);
    lHelper.RichView.Style := lRVStyle;
    FStream.Position := 0;
    lHelper.RichView.LoadRVFFromStream(FStream);
    lHelper.Init(FCanvasRef, FPageWidth);
    lHelper.RichView.RTFOptions := [rvrtfDuplicateUnicode, rvrtfSaveEMFAsWMF,
      rvrtfSaveBitmapDefault];
    lHelper.RichView.SaveRTFToStream(AStream, False);
  finally
    lHelper.Free;
    lRVStyle.Free;
  end;
end;

{============== TraTppCustomRichViewRTTI  (TraTppStretchableRTTI) =============}
class function TraTppCustomRichViewRTTI.RefClass: TClass;
begin
  Result := TppCustomRichView;
end;

{------------------------------------------------------------------------------}
class procedure TraTppCustomRichViewRTTI.GetPropList(aClass: TClass;
  aPropList: TraPropList);
begin
  inherited GetPropList(aClass, aPropList);
  aPropList.AddProp('RVReportHelper');
  aPropList.AddMethod('LoadRTFFromFile');
  aPropList.AddMethod('LoadRTFFromStream');
  aPropList.AddMethod('Clear');
  aPropList.AddMethod('SaveRTFToFile');
  aPropList.AddMethod('SaveRTFToStream');
end;

{------------------------------------------------------------------------------}
class function TraTppCustomRichViewRTTI.GetPropRec(aClass: TClass;
  const aPropName: String; var aPropRec: TraPropRec): Boolean;
begin
  Result := True;
  if (CompareText(aPropName, 'LoadRTFFromFile') = 0) or
    (CompareText(aPropName, 'LoadRTFFromStream') = 0) or
    (CompareText(aPropName, 'Clear') = 0) or
    (CompareText(aPropName, 'SaveRTFToFile') = 0) or
    (CompareText(aPropName, 'SaveRTFToStream') = 0) then
    MethodToRec(aPropName, True, aPropRec)
  else if (CompareText(aPropName, 'RVReportHelper') = 0) then
    PropToRec(aPropName, daClass, False, aPropRec)
  else
    Result := inherited GetPropRec(aClass, aPropName, aPropRec);
end;

{------------------------------------------------------------------------------}
class function TraTppCustomRichViewRTTI.GetPropValue(aObject: TObject;
  const aPropName: String; var aValue): Boolean;
begin
  Result := True;
  if (CompareText(aPropName, 'RVReportHelper') = 0) then
    TObject(aValue) := TppCustomRichView(aObject).Helper
  else
    Result := inherited GetPropValue(aObject, aPropName, aValue);
end;

{------------------------------------------------------------------------------}
class function TraTppCustomRichViewRTTI.SetPropValue(aObject: TObject;
  const aPropName: String; var aValue): Boolean;
begin
  if (CompareText(aPropName, 'RVReportHelper') = 0) then
    Result := True
  else
    Result := inherited SetPropValue(aObject, aPropName, aValue);
end;

{------------------------------------------------------------------------------}
class function TraTppCustomRichViewRTTI.CallMethod(aObject: TObject;
  const aMethodName: String; aParams: TraParamList; aGet: Boolean): Boolean;
var
  lRichView: TppCustomRichView;
  LStream:   TStream;
  lString:   String;
begin
  Result := True;
  lRichView := TppCustomRichView(aObject);
  if ppEqual(aMethodName, 'LoadRTFFromStream') then
  begin
    aParams.GetParamValue(0, LStream);
    lRichView.RichView.LoadRTFFromStream(LStream);
    lRichView.Reset;
    lRichView.PropertyChange;
  end
  else if ppEqual(aMethodName, 'LoadRTFFromFile') then
  begin
    aParams.GetParamValue(0, lString);
    lRichView.RichView.LoadRTF(lString);
    lRichView.Reset;
    lRichView.PropertyChange;
  end
  else if ppEqual(aMethodName, 'SaveToRTFStream') then
  begin
    aParams.GetParamValue(0, LStream);
    lRichView.RichView.SaveRTFToStream(LStream, False);
  end
  else if ppEqual(aMethodName, 'Clear') then
  begin
    lRichView.RichView.Clear;
  end
  else if ppEqual(aMethodName, 'SaveRTFToFile') then
  begin
    aParams.GetParamValue(0, lString);
    lRichView.RichView.SaveRTF(lString, False);
  end
  else if ppEqual(aMethodName, 'AfterEdit') then
  begin
    lRichView.Reset;
    lRichView.PropertyChange;
    lRichView.InvalidateDesignControl;
  end
  else
    Result := inherited CallMethod(aObject, aMethodName, aParams, aGet);
end;

{------------------------------------------------------------------------------}
class function TraTppCustomRichViewRTTI.GetParams(const aMethodName: String)
  : TraParamList;
begin
  if ppEqual(aMethodName, 'LoadRTFFromFile') then
  begin
    Result := TraParamList.Create;
    Result.AddParam('aFileName', daString, nil, '', False, False);
  end
  else if ppEqual(aMethodName, 'LoadRTFFromStream') then
  begin
    Result := TraParamList.Create;
    Result.AddParam('aStream', daClass, TStream, '', False, False);
  end
  else if ppEqual(aMethodName, 'SaveRTFToFile') then
  begin
    Result := TraParamList.Create;
    Result.AddParam('aFileName', daString, nil, '', False, False);
  end
  else if ppEqual(aMethodName, 'SaveRTFToStream') then
  begin
    Result := TraParamList.Create;
    Result.AddParam('aStream', daClass, TStream, '', False, False);
  end
  else
    Result := inherited GetParams(aMethodName);
end;

{==============================================================================}

initialization

ppRegisterComponent(TppRichView, 'Standard Components', 12, 0,
  'Advanced RichText', HInstance);
ppRegisterComponent(TppDBRichView, 'Data Components', 12, 0,
  'Advanced DBRichText', HInstance);

raRegisterRTTI(TraTppCustomRichViewRTTI);
RegisterClasses([TppDrawRichView]);

finalization

ppUnRegisterComponent(TppRichView);
ppUnRegisterComponent(TppDBRichView);
UnRegisterClasses([TppDrawRichView]);
raUnRegisterRTTI(TraTppCustomRichViewRTTI);

end.
