{*******************************************************}
{                                                       }
{       ReportBuilder Wrapper for RichView Package      }
{       Default editor for ppRichView                   }
{                                                       }
{       ReportBuilder                                   }
{       Copyright (c) Digital Metaphors Corporation     }
{                                                       }
{       RichView                                        }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

unit ppRichViewEd;

interface

{$J+}
{$B-}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ppForms, RVStyle, StdCtrls, ExtCtrls, RVScroll, RichView, RVEdit,
  RVOfficeCnv, ppRichViewCustomEd;

type
  TfrmppRichViewEditor = class(TfrmCustomppRichViewEditor)
    rv: TRichView;
    Panel1: TPanel;
    od: TOpenDialog;
    btnOpen: TButton;
    btnOk: TButton;
    btnCancel: TButton;
    RVStyle1: TRVStyle;
    btnClear: TButton;
    rvoc: TRVOfficeConverter;
    btnPaste: TButton;
    rve: TRichViewEdit;
    procedure btnOpenClick(Sender: TObject);
    procedure btnClearClick(Sender: TObject);
    procedure btnPasteClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
  public
    function GetRichView: TCustomRichView; override;
  end;

implementation

{$R *.DFM}

{------------------------------------------------------------------------------}
procedure TfrmppRichViewEditor.btnOpenClick(Sender: TObject);
var
  r: Boolean;
begin
  if od.Execute then
  begin
    Screen.Cursor := crHourGlass;
    try
      rv.Clear;
      rv.DeleteUnusedStyles(True, True, True);
      case od.FilterIndex of
        1:
          r := rv.LoadRTF(od.FileName);
        2:
          r := rv.LoadText(od.FileName, 0, 0, False);
        3:
          r := rv.LoadRVF(od.FileName);
        else
          r := rvoc.ImportRV(od.FileName, rv, od.FilterIndex - 4)
      end;
      rv.Format;
    finally
      Screen.Cursor := crDefault;
    end;
    if not r then
      if od.FilterIndex < 4 then
        Application.MessageBox('Error loading file', 'Error', MB_OK)
      else
        Application.MessageBox
          ('Microsoft Office converter reports error when importing the file ',
          'Error', MB_OK);
  end;
end;

{------------------------------------------------------------------------------}
procedure TfrmppRichViewEditor.btnClearClick(Sender: TObject);
begin
  rv.Clear;
  rv.LeftMargin := 5;
  rv.RightMargin := 5;
  rv.TopMargin := 5;
  rv.BottomMargin := 5;
  rv.Color := clWindow;
  rv.BackgroundStyle := bsNoBitmap;
  rv.Format;
end;

{------------------------------------------------------------------------------}
procedure TfrmppRichViewEditor.btnPasteClick(Sender: TObject);
var
  Stream: TMemoryStream;
begin
  if rve.CanPaste then
  begin
    Screen.Cursor := crHourGlass;
    try
      rve.Paste;
      Stream := TMemoryStream.Create;
      rve.SaveRVFToStream(Stream, False);
      Stream.Position := 0;
      rv.Clear;
      rv.LoadRVFFromStream(Stream);
      rv.Format;
      rve.Clear;
    finally
      Screen.Cursor := crDefault;
    end;
  end
  else
    Application.MessageBox
      ('Nothing to paste: the Clipboard does not contain supported formats',
      'Pasting', MB_ICONEXCLAMATION);
end;

{------------------------------------------------------------------------------}
procedure TfrmppRichViewEditor.FormActivate(Sender: TObject);
begin
  rve.RVFTextStylesReadMode := rvf_sInsertMerge;
  rve.RVFParaStylesReadMode := rvf_sInsertMerge;
  rve.RVFOptions := rv.RVFOptions;
  rve.Options := rv.Options;
  rve.RTFReadProperties.TextStyleMode := rvrsAddIfNeeded;
  rve.RTFReadProperties.ParaStyleMode := rvrsAddIfNeeded;
  rve.BidiMode := rv.BiDiMode;
end;

{------------------------------------------------------------------------------}
procedure TfrmppRichViewEditor.FormCreate(Sender: TObject);
begin
  od.Filter := od.Filter + '|' + rvoc.GetImportFilter;
end;

{------------------------------------------------------------------------------}
function TfrmppRichViewEditor.GetRichView: TCustomRichView;
begin
  Result := rv;
end;

initialization

ppRichViewEditorClass := TfrmppRichViewEditor;

end.
