{*******************************************************}
{                                                       }
{       ReportBuilder Wrapper for RichView Package      }
{       Report Designer Support                         }
{                                                       }
{       ReportBuilder                                   }
{       Copyright (c) Digital Metaphors Corporation     }
{                                                       }
{       RichView                                        }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

{$J+}
{$B-}
{$I RV_Defs.inc}
unit ppRichViewReportDesign;

interface

uses

  Windows,
  SysUtils, Classes, Graphics, Controls, Forms, StdCtrls, ExtCtrls, Menus,
  ComCtrls, Dialogs,
  ppTypes,
  ppPopupMenus,
  ppDesignControls,
  ppDesignControlsEx,
  ppRichView,
  CRVData, RVClasses, RVRVData, RVFuncs, RichEdit, RichView, RVStyle, RVReport,
  PtblRV,
  RVStr;

type

  {Popup menu classes used to display the TRichView
   design control's context menu in the report designer}

  TppRichViewCustomPopupMenu = class(TppStretchablePopupMenu)
  protected
    procedure CreateMenuItems; override;
  end;

  TppRichViewPopupMenu = class(TppRichViewCustomPopupMenu)
  private
    procedure EditMenuClick(Sender: TObject);
  protected
    procedure CreateMenuItems; override;
    procedure DoPopup(Sender: TObject); override;
  end;

  TppDBRichViewPopupMenu = class(TppRichViewCustomPopupMenu)
  end;

  {Design control classes used to display the TRichView
   in the report designer}

  TppRichViewCustomControl = class(TppCustomComponentControl)
  private
    FMetafile: TMetafile;
    procedure MakeDesignTimeMetafile(Canvas: TCanvas; const Rect: TRect;
      rvrh: TRVReportHelper);
  protected
    procedure PaintDesignControl(aCanvas: TCanvas); override;
  public
    destructor Destroy; override;
  end;

  TppRichViewControl = class(TppRichViewCustomControl)
  end;

  TppDBRichViewControl = class(TppRichViewCustomControl)
  end;

  {declare Register procedure so component can be used at Delphi design-time}
procedure Register;

implementation

uses
  {$IFDEF RICHVIEWDEFXE2}
  UITypes,
  {$ENDIF}
  ppRichViewCustomEd, ppRichViewEd;

{============================== TReportRVDataShared ===========================}
{
 TReportRVDataShared - TRVReportHelper sharing items with external RichView
}

type
  TReportRVDataShared = class(TReportRVData)
  protected
    function ShareItems: Boolean; override;
  end;

  TReportRichViewShared = class(TReportRichView)
  protected
    function GetDataClass: TRichViewRVDataClass; override;
  end;

  TRVReportHelperShared = class(TRVReportHelper)
  protected
    function CreateRichView: TCustomPrintableRV; override;
    function GetSourceRichView: TCustomRichView; override;
    {$IFNDEF RVDONOTUSESEQ}
    function IgnoreEndnotes: Boolean; override;
    {$ENDIF}
  public
    SourceRichView: TCustomRichView;
  end;

  {------------------------------------------------------------------------------}
function TReportRVDataShared.ShareItems: Boolean;
begin
  Result := True;
end;

{------------------------------------------------------------------------------}
function TReportRichViewShared.GetDataClass: TRichViewRVDataClass;
begin
  Result := TReportRVDataShared;
end;

{------------------------------------------------------------------------------}
function TRVReportHelperShared.CreateRichView: TCustomPrintableRV;
begin
  Result := TReportRichViewShared.Create(Self);
end;

{------------------------------------------------------------------------------}
function TRVReportHelperShared.GetSourceRichView: TCustomRichView;
begin
  Result := SourceRichView;
end;
{------------------------------------------------------------------------------}
{$IFNDEF RVDONOTUSESEQ}

function TRVReportHelperShared.IgnoreEndnotes: Boolean;
begin
  Result := True;
end;
{$ENDIF}

{============================== TppRichViewCustomPopupMenu ====================}
procedure TppRichViewCustomPopupMenu.CreateMenuItems;
begin
  inherited;
  AddItem(30, 'Transparent', '', 246);
  AddItem(30, 'KeepTogether', '', 1130);
end;

{=============================== TppRichViewPopupMenu =========================}
procedure TppRichViewPopupMenu.CreateMenuItems;
begin
  inherited;
  AddItem(30, 'Edit', '', 120);
end;

{------------------------------------------------------------------------------}
procedure TppRichViewPopupMenu.DoPopup(Sender: TObject);
begin
  inherited;
  ItemByName('Edit').OnClick := EditMenuClick;
end;

{------------------------------------------------------------------------------}
procedure TppRichViewPopupMenu.EditMenuClick(Sender: TObject);
var
  frm:    TfrmCustomppRichViewEditor;
  Stream: TMemoryStream;
  formrv: TCustomRichView;
  rvrh:   TRVReportHelper;
  Color:  TColor;

  function CheckRichViewVersion: Boolean;
  var
    p, v: Integer;
    s:    String;
  begin
    Result := False;
    s := Copy(RVVersion, 2, Length(RVVersion));
    p := Pos('.', s);
    v := StrToIntDef(Copy(s, 1, p - 1), 0);
    if v > 1 then
    begin
      Result := True;
      exit;
    end;
    s := Copy(s, p + 1, Length(s));
    p := Pos('.', s);
    if p = 0 then
      p := Length(s) + 1;
    v := StrToIntDef(Copy(s, 1, p - 1), 0);
    if v <> 7 then
    begin
      Result := v > 7;
      exit;
    end;
    if p = Length(s) + 1 then
      exit;
    s := Copy(s, p + 1, Length(s));
    v := StrToIntDef(s, 0);
    Result := v > 4;
  end;

begin

  if (csDesigning in ComponentState) and not CheckRichViewVersion then
  begin
    MessageBox(0,
      'This command is available at design-time for RichView 1.7.5 or newer',
      'Run-time only', MB_ICONSTOP);
    exit;
  end;

  frm := ppRichViewEditorClass.Create(nil);
  formrv := frm.GetRichView;

  GetPropValue('RVReportHelper', rvrh);
  GetPropValue('Color', Color);
  formrv.AssignEvents(rvrh.RichView);
  rvrh.RichView.Color := Color;

  with formrv do
  begin
    RVFTextStylesReadMode := rvf_sInsertMerge;
    RVFParaStylesReadMode := rvf_sInsertMerge;
    RVFOptions := rvrh.RichView.RVFOptions;
    Options := rvrh.RichView.Options;
    RTFReadProperties.TextStyleMode := rvrsAddIfNeeded;
    RTFReadProperties.ParaStyleMode := rvrsAddIfNeeded;
    RTFReadProperties.UnicodeMode :=
      rvrh.RichView.RTFReadProperties.UnicodeMode;
    Color := rvrh.RichView.Color;
    BackgroundBitmap := rvrh.RichView.BackgroundBitmap;
    BackgroundStyle := rvrh.RichView.BackgroundStyle;
    LeftMargin := rvrh.RichView.LeftMargin;
    RightMargin := rvrh.RichView.RightMargin;
    TopMargin := rvrh.RichView.TopMargin;
    BottomMargin := rvrh.RichView.BottomMargin;
    BiDiMode := rvrh.RichView.BiDiMode;
  end;

  Screen.Cursor := crHourGlass;
  try
    Stream := TMemoryStream.Create;
    rvrh.RichView.SaveRVFToStream(Stream, False);
    Stream.Position := 0;
    formrv.LoadRVFFromStream(Stream);
    Stream.Free;
    formrv.Format;
  finally
    Screen.Cursor := crDefault;
  end;
  if frm.ShowModal = mrOk then
  begin
    Screen.Cursor := crHourGlass;
    try
      Stream := TMemoryStream.Create;
      formrv.SaveRVFToStream(Stream, False);
      Stream.Position := 0;
      rvrh.RichView.Clear;
      rvrh.RichView.DeleteUnusedStyles(True, True, True);
      rvrh.RichView.LoadRVFFromStream(Stream);
      Stream.Free;
    finally
      Screen.Cursor := crDefault;
    end;
    Color := formrv.Color;
    SetPropValue('Color', Color);
    GetPropValue('RVReportHelper', rvrh);
    GetRTTIClass.CallMethod(Component, 'AfterEdit', nil, True);
  end;
  frm.Free;
end;

{================================ TppRichViewCustomControl ====================}
destructor TppRichViewCustomControl.Destroy;
begin
  FMetafile.Free;
  inherited;
end;

{------------------------------------------------------------------------------}
procedure TppRichViewCustomControl.PaintDesignControl(aCanvas: TCanvas);
var
  lSaveColor:    TColor;
  lClientRect:   TRect;
  lbTransparent: Boolean;
  rvrh:          TRVReportHelper;
begin
  lClientRect := ClientRect;
  GetPropValue('Transparent', lbTransparent);
  GetPropValue('RVReportHelper', rvrh);

  if not lbTransparent then
  begin
    {draw background color}
    aCanvas.Brush.Color := GetColor;
    aCanvas.Brush.Style := bsSolid;
    aCanvas.FillRect(lClientRect);
  end;

  if rvrh.RichView.ItemCount > 0 then
  begin
    MakeDesignTimeMetafile(aCanvas, lClientRect, rvrh);
    aCanvas.Draw(lClientRect.Left + 1, lClientRect.Top + 1, FMetafile);
  end
  else
  begin
    aCanvas.Brush.Style := bsClear;
    with aCanvas.Font do
    begin
      Size := 8;
      Name := 'MS Sans Serif';
      Color := clWindowText;
      Style := [];
      Charset := DEFAULT_CHARSET;
    end;
    aCanvas.TextRect(lClientRect, lClientRect.Left + 1,
      lClientRect.Top + 1, Caption);
  end;

  {draw a framerect}
  lSaveColor := aCanvas.Brush.Color;
  aCanvas.Brush.Style := bsSolid;
  aCanvas.Brush.Color := clBlack;
  aCanvas.FrameRect(lClientRect);
  aCanvas.Brush.Color := lSaveColor;
end;

{------------------------------------------------------------------------------}
procedure TppRichViewCustomControl.MakeDesignTimeMetafile(Canvas: TCanvas;
  const Rect: TRect; rvrh: TRVReportHelper);
var
  tmphelper:       TRVReportHelperShared;
  lMetafileCanvas: TMetafileCanvas;
  lCursor:         TCursor;
  lbTransparent:   Boolean;
begin
  lCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  GetPropValue('Transparent', lbTransparent);
  try
    tmphelper := TRVReportHelperShared.Create(nil);
    try
      rvrh.RichView.Style.TextStyles.PixelsPerInch := 0;
      tmphelper.RichView.Style := rvrh.RichView.Style;
      tmphelper.RichView.Options := rvrh.RichView.Options;
      tmphelper.RichView.RVData.ShareItemsFrom(rvrh.RichView.RVData);
      tmphelper.SourceRichView := rvrh.RichView;
      tmphelper.RichView.BackgroundBitmap := rvrh.RichView.BackgroundBitmap;
      tmphelper.RichView.BackgroundStyle := rvrh.RichView.BackgroundStyle;
      tmphelper.RichView.LeftMargin := rvrh.RichView.LeftMargin;
      tmphelper.RichView.RightMargin := rvrh.RichView.RightMargin;
      tmphelper.RichView.TopMargin := rvrh.RichView.TopMargin;
      tmphelper.RichView.BottomMargin := rvrh.RichView.BottomMargin;
      tmphelper.RichView.BiDiMode := rvrh.RichView.BiDiMode;
      tmphelper.RichView.AssignEvents(rvrh.RichView);
      // tmphelper.RichView.HandleNeeded;
      tmphelper.Init(Canvas, Rect.Right - Rect.Left - 1);
      tmphelper.FormatNextPage(Rect.Bottom - Rect.Top - 1);
      FMetafile.Free;
      FMetafile := TMetafile.Create;
      FMetafile.Width := Rect.Right - Rect.Left - 1;
      FMetafile.Height := Rect.Bottom - Rect.Top - 1;
      tmphelper.TransparentBackground := lbTransparent;
      if not lbTransparent then
        tmphelper.RichView.Color := GetColor;
      lMetafileCanvas := TMetafileCanvas.Create(FMetafile, 0);
      try
        if tmphelper.PagesCount > 0 then
          tmphelper.DrawPage(1, lMetafileCanvas, True,
            Rect.Bottom - Rect.Top - 1);
      finally
        lMetafileCanvas.Free;
      end;
    finally
      rvrh.RichView.Style.TextStyles.PixelsPerInch := TPPRICHVIEWREFCANVASPPI;
      tmphelper.Free;
    end;
  finally
    Screen.Cursor := lCursor;
  end;
end;

{==============================================================================}
procedure Register;
begin
  {register the component for use in the Delphi design-time report designer}
  RegisterNoIcon([TppRichView]);
  RegisterNoIcon([TppDBRichView]);
end;

initialization

{Register Popup Menus}
TppPopupMenuManager.RegisterMenuClass(TppRichView, TppRichViewPopupMenu);
TppPopupMenuManager.RegisterMenuClass(TppDBRichView, TppDBRichViewPopupMenu);

{Register Design Controls}
TppDesignControlFactory.RegisterDesignControlClass(TppRichView,
  TppRichViewControl);
TppDesignControlFactory.RegisterDesignControlClass(TppDBRichView,
  TppDBRichViewControl);

end.
